require 'resource_manager/version'

module ResourceManager
  class Base

    class_attribute :before_create_callback, :after_create_callback, :before_update_callback, :after_update_callback, :before_destroy_callback, :after_destroy_callback
    class_attribute :create_action, :update_action, :destroy_action

    class << self
      def model(class_or_proc, &block)
        @model = class_or_proc
      end

      [:before_create, :after_create, :before_update, :after_update, :before_destroy, :after_destroy].each do |method|
        define_method(method) do |&block|
          self.send("#{method}_callback=", block)
        end
      end

      [:create, :update, :destroy].each do |method|
        define_method(method) do |&block|
          self.send("#{method}_action=", block)
        end
      end
    end

    def model
      model_or_proc = self.class.instance_variable_get(:@model)
      if model_or_proc.kind_of?(Proc)
        instance_exec(&model_or_proc)
      else
        model_or_proc
      end
    end

    ['', '!'].each do |name|
      define_method "create#{name}" do |params, &block|
        transaction do
          run_callback(:before_create, params)

          if self.create_action.present?
            instance_exec(params, &self.create_action)
          else
            item = model.new(params)
            saved = item.send("save#{name}")
            block.call(item, saved) if block.present?
            item
          end

          run_callback(:after_create, item, params) if saved
          item
        end
      end

      define_method "update#{name}" do |item_or_id, params, &block|
        transaction do
          item = item(item_or_id)
          run_callback(:before_update, item, params)

          if self.update_action.present?
            instance_exec(item, params, &self.update_action)
          else
            item.assign_attributes(params)
            changed = item.changed
            saved = item.send("save#{name}")
            block.call(item, saved) if block.present?
            item
          end

          run_callback(:after_update, item, params, changed) if saved
          item
        end
      end

      define_method "destroy#{name}" do |item_or_id|
        transaction do
          item = item(item_or_id)
          run_callback(:before_destroy, item)

          if self.destroy_action.present?
            instance_exec(item, &self.destroy_action)
          else
            item.send("destroy#{name}")
          end

          run_callback(:after_destroy, item)
          item
        end
      end
    end

    private

    def run_callback(name, *args)
      self.instance_exec(*args, &self.send("#{name}_callback")) if self.send("#{name}_callback").present?
    end

    def item(item_or_id)
      if item_or_id.is_a?(Numeric)
        model.find(item_or_id)
      else
        item_or_id
      end
    end

    def transaction(&block)
      ActiveRecord::Base.transaction &block
    end
  end
end
